/*	
  Copyright(c) 2011 Christoph Schmalz <christoph.schmalz@informatik.uni-erlangen.de>
  
  This file is licensed under the terms of of the GNU General Public License, 
  version 3, as published by the Free Software Foundation. You can find it at:
  http://www.gnu.org/licenses/gpl.html
*/

#include "stdafx.h"
#include "util.h"

///////////////////////////////////////////////////////////////////////////////
// Function: EvaluatePhaseShift
// Description:
//	 Three or four step phase shift
///////////////////////////////////////////////////////////////////////////////

void EvaluatePhaseShift(
	const imageArray_t&		images,			//in: three or four images
	int						minDynamic,		//in: minimal dynamic to compute phase
	Mat*					pWrappedPhase,	//in: receives the wrapped phase
	Mat*					pMaskImg,		//in: receives mask
	int						iOverExpThresh) //in: threshold for overexposure, typically 250
{
	// Sanity check
	if (!pWrappedPhase || !pMaskImg || images.size()<3)
		return;

	pWrappedPhase->create( images[0]->size(), CV_32FC1 );
	pMaskImg->create( images[0]->size(), CV_8UC1 );

	const int				N = (int)images.size();
	std::vector<uint8_t*>	data(N,0);
	imageArray_t			gray(N,0);
	std::vector<float>		sin_lut(N),cos_lut(N);

	//prepare
	for (int i=0; i<N; i++)
	{
		sin_lut[i] = sin(M_2PIf/N*i);
		cos_lut[i] = cos(M_2PIf/N*i);

		if (images[i]->type() != CV_8UC1 )
			return;

		data[i] = (uint8_t*)images[i]->data;
	}	

	//the only important thing
	float*		pOut	= (float*)pWrappedPhase->data;
	uint8_t*	pui8Mask= (uint8_t*)pMaskImg->data;

	#pragma omp parallel
	{
		std::vector<uint8_t> v(N,0);
		#pragma omp for
		for (size_t i=0; i<pWrappedPhase->total(); i++)
		{
			int iOverExpCount=0;
			for (int j=0; j<N; j++)
			{
				v[j] = data[j][i];
				if (v[j]>iOverExpThresh)
					iOverExpCount++;
			}

			if (iOverExpCount==0)
			{
				//here the magic happens
				float a=0,b=0;
				for (int j=0; j<N; j++)
				{
					a += v[j]*sin_lut[j];
					b += v[j]*cos_lut[j];
				}

				float amplitude = sqrt( a*a+b*b )*2/N;
				if (amplitude>=minDynamic)
				{
					//phase
					pOut[i] = atan2( a,b );
					if (pOut[i]<0) 
						pOut[i]+=M_2PIf;
					//modulation
					pui8Mask[i] = (uint8_t)amplitude;
				}
				else
					pOut[i] = pui8Mask[i] = 0;
			}
			else
				pOut[i] = pui8Mask[i] = 0;
		}
	}

	// cleanup
	for (int i=0; i<N; i++)
		delete gray[i];		
}


///////////////////////////////////////////////////////////////////////////////
// Function: UnwrapPhaseWithGrayCode
// Description:
//	unwrap phase with gray code stripe numbers.
//	phaseshift wavelength is assumed to be twice the gc stripe width
//	to solve cases of misalignment
///////////////////////////////////////////////////////////////////////////////

void UnwrapPhaseWithGrayCode(
 Mat*		pPhaseImg,		//in: phase image of type float
 Mat*		pQImg,			//in: quality image, may be NULL
 Mat*		pIndexImg,		//in: index image with integer stripe ids
 float		fNrOfGCStripes,	//in: total nr of gray code stripes
 uint8_t	ucMinQuality,	//in: minimal quality
 bool		bNormalizePhase)//in: normalize unwrapped phase to [0:1]?
{
	if (!pPhaseImg || !pQImg || !pIndexImg )
		return;

	uint32_t*	piStripeNr	= (uint32_t*)pIndexImg->data;
	float*		pfPhase		= (float*)pPhaseImg->data;
	uint8_t*	pucMask		= (uint8_t*)pQImg->data;
	const int	iWidth		= pIndexImg->size().width;
	const int	iHeight		= pIndexImg->size().height;
	const float fPi1pt5		= M_PIf*1.5f;
	const float fPi0pt5		= M_PIf*0.5f;

	//we may want to normalize the result to one
	const float	fNormalizationFactor =
		bNormalizePhase ? 1 / (M_2PIf * fNrOfGCStripes ) : 1;
	assert(_finite(fNormalizationFactor));	
	
#pragma omp parallel for
	for (int iRow=0; iRow<iHeight; iRow++)
	{
		for (int iCol=0; iCol<iWidth; iCol++)
		{
			int		iIndex		= iCol + iRow*iWidth;
			int		iGCStripeId = piStripeNr[iIndex];
			float	fPhase		= pfPhase[iIndex];
			uint8_t	ucMask		= pucMask[iIndex];

			// Consider this pixel at all?
			if (ucMask<ucMinQuality) 
			{
				pfPhase[iIndex]		= 0;
				piStripeNr[iIndex]	= 0;
				continue;
			}

			// To counter misalignment, each phase interval stretches over two 
			// gray code stripes. The stripe ids start at zero. So we know that 
			// values < pi should be on an even stripe and values > pi should
			// be on an odd stripe

			// Fix the ids heuristically
			// even stripe but phase is high?
			if ( !( iGCStripeId&1 ) && ( fPhase>M_PIf ) ) 
			{
				if ( fPhase<fPi1pt5 )
				//beginning is off. shift it to the next stripe
					iGCStripeId++;
				else
				//end is off. shift it to the previous stripe
					iGCStripeId--;
			}
			//odd stripe but phase is low?
			else if ( ( iGCStripeId&1 ) && ( fPhase<M_PIf ) ) 
			{
				if ( fPhase<fPi0pt5)
				//beginning is off. shift it to the next stripe
					iGCStripeId++;
				else
				//end is off. shift it to the previous stripe
					iGCStripeId--;
			}

			// Now fix the phase so it is in [0;2pi[ on
			// each single stripe. Mind the >=!
			if (fPhase>=M_PIf)
				fPhase-=M_PIf;
			fPhase*=2;

			// unwrap, normalize and store
			pfPhase[iIndex]	= (iGCStripeId*M_2PIf + fPhase) * fNormalizationFactor;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Function: ClassifyGraycodeImg
// FunctionType: []
// Description:
//   Classify all pixels of an image and set the current bit according to the
//   result of the classification.
//   It is assumed that all bits of the result image are 0 per default.
///////////////////////////////////////////////////////////////////////////////

bool // true if successful, false if an error occured.
 ClassifyGraycodeImg
  (Mat*	pPosImg,	//In: img taken under the Gray code pattern illu
   Mat*	pNegImg,	//In: img taken under inverse Gray code pattern illu
   int	iThres,		//In: Threshold specifying minimal intensity variation
   Mat*	pQImg,		//In: img for contrast (can be NULL)
   Mat*	pResultImg,	//In: img that receives result
   int	iCurBit)	//In: bit to set in the result image
{
	// All images have to be of the same type
	if ( !pResultImg ||
		pPosImg->type() != CV_8UC1 ||
		pNegImg->type() != CV_8UC1 ||
		pPosImg->size() != pNegImg->size() )
		return false;

	const int	iWidth		= pPosImg->size().width;
	const int	iHeight		= pPosImg->size().height;

	uint32_t*	pResult		= (uint32_t*)pResultImg->data;	
	uint8_t*	pQ			= (uint8_t*)pQImg ? pQImg->data : 0;

	uint8_t*	pucPos		= (uint8_t*)pPosImg->data;
	uint8_t*	pucNeg		= (uint8_t*)pNegImg->data;

#pragma omp parallel for
	for (int iRow=0; iRow<iHeight; iRow++)
		for (int iCol=0; iCol<iWidth; iCol++)
		{
			int	iIndex		= iCol+iRow*iWidth;
			int iPos		= pucPos[iIndex];
			int iNeg		= pucNeg[iIndex];					
			int iDeltaAbs	= abs(iPos-iNeg);

			if (pQ) pQ[iIndex]	= iDeltaAbs;

			// Compare the two values - there should be a noticable difference
			if (iDeltaAbs<iThres)
			{
				// Bad pixel
				pResult[iIndex]	= 0;
				continue;
			}

			/*
			// Filter out highlights ...
			if (iPos==255 || iNeg==255)
			{
				pResult[iIndex]	= 0;
				continue;
			}
			*/

			// Finally, write out the bit (nothing to do in the 0-case)
			if (iPos>iNeg)
				pResult[iIndex]+=iCurBit;
		}
	

	return true;
}


///////////////////////////////////////////////////////////////////////////////
// Function: ConvertGrayCodeToBinary
// FunctionType: []
// Description:
//
///////////////////////////////////////////////////////////////////////////////
void ConvertGrayCodeToBinary(Mat* img)
{
	const size_t	iWidth		= img->size().width;
	const size_t	iHeight		= img->size().height;
	uint32_t*		puiResult	= (uint32_t*)img->data;

	#pragma omp parallel for
	for (size_t iRow=0; iRow<iHeight; iRow++)
		for (size_t iCol=0;iCol<iWidth; iCol++)
		{
			const int	 iIndex			= iCol+iRow*iWidth;
			uint32_t	uiGraycode		= puiResult[iIndex];
			uint32_t	uiBinaryCode	= GrayCodeToBinary(uiGraycode);
			puiResult[iIndex] = uiBinaryCode; 
		}

	//we could potentially do more here, for example
	//filter bad pixels, interpolate holes and so on
}